/*
 * h264codec.cxx   
 *
 * H.323 protocol handler
 *
 * Open H323 Library
 * 
 */

/*
 *  Initial release notes from Marek Domaracky::
 *
 * Added support for video capabilities & codec.
 * The code for is copied from h263codec.cxx,
 *
 */
#include <ptlib.h>

#ifdef __GNUC__
#pragma implementation "h264codec.h"
#endif

#include "h264codec.h"
#include "codec/opalplugin.h"


#include "h245.h"
#include "rtp.h"

#if defined(_MSC_VER)
#pragma comment(lib, H323_VICH263_LIBRARY)
#endif


#define new PNEW

char OpalH264[] = "H.264";
OPAL_MEDIA_FORMAT_DECLARE(OpalH264Format,
        OpalH264,
        OpalMediaFormat::DefaultVideoSessionID,
        RTP_DataFrame::DynamicBase,
        FALSE,  // No jitter for video
        180000, // bits/sec
        2000,   // Not sure of this value!
        0,      // No intrinsic time per frame
        OpalMediaFormat::VideoTimeUnits,
        0)

//////////////////////////////////////////////////////////////////////////////

H323_H264Capability::H323_H264Capability(unsigned _sqcifMPI,
					 unsigned _qcifMPI,
					 unsigned _cifMPI,
					 unsigned _cif4MPI,
					 unsigned _cif16MPI,
					 unsigned _maxBitRate,
					 unsigned _videoFrameRate,
					 PBoolean _unrestrictedVector,
					 PBoolean _arithmeticCoding,
					 PBoolean _advancedPrediction,
					 PBoolean _pbFrames
					)
{
  

  maxBitRate = _maxBitRate;
 
  pbFrames = _pbFrames;
 

 
}


PObject * H323_H264Capability::Clone() const
{
return new H323_H264Capability(*this);
}


PObject::Comparison H323_H264Capability::Compare(const PObject & obj) const
{
 if (!PIsDescendant(&obj, H323_H264Capability))
    return LessThan;

  Comparison result = H323Capability::Compare(obj);
  if (result != EqualTo) 
    return result;

  const H323_H264Capability & other = (const H323_H264Capability &)obj;



    return EqualTo;

  if (qcifMPI > 0)
    return LessThan;

  return GreaterThan;


  return EqualTo;
}


PString H323_H264Capability::GetFormatName() const
{
PString ret = "H.264";// H263_MediaFormat;
  return ret;
}


unsigned H323_H264Capability::GetSubType() const
{
return H245_VideoCapability::e_genericVideoCapability;
}


PBoolean H323_H264Capability::OnSendingPDU(H245_VideoCapability & cap) const
{
    cap.SetTag(H245_VideoCapability::e_genericVideoCapability);

    H245_GenericCapability & h264 = cap;
    h264.m_capabilityIdentifier = *(new H245_CapabilityIdentifier(H245_CapabilityIdentifier::e_standard));
    PASN_ObjectId &object_id = h264.m_capabilityIdentifier;
    object_id = OpalPluginCodec_Identifer_H264_Generic;
    h264.IncludeOptionalField(H245_GenericCapability::e_maxBitRate);
    h264.m_maxBitRate = maxBitRate;
    h264.IncludeOptionalField(H245_GenericCapability::e_collapsing);
    
    H245_GenericParameter * param = new H245_GenericParameter;

    param->m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
    (PASN_Integer &)param->m_parameterIdentifier = 41;
    unsigned parameterValue = 41; // H.264 Profile parameter
    unsigned tag = H245_ParameterValue::e_booleanArray;
        param->m_parameterValue.SetTag(tag);
      (PASN_Integer &)param->m_parameterValue = 64; // Main profile
       h264.m_collapsing.Append(param); 
       
        H245_GenericParameter * param1 = new H245_GenericParameter;
    param1->m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
       (PASN_Integer &)param1->m_parameterIdentifier = 42; //H.264 Level parameter
       tag = H245_ParameterValue::e_unsignedMin;
        param1->m_parameterValue.SetTag(tag);
      (PASN_Integer &)param1->m_parameterValue = 57; // level 4
       h264.m_collapsing.Append(param1);
       
        H245_GenericParameter * param2 = new H245_GenericParameter;
    param2->m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
       (PASN_Integer &)param2->m_parameterIdentifier = 3; //H.264 CustomMaxMBPS processing rate parameter
       tag = H245_ParameterValue::e_unsignedMin;
        param2->m_parameterValue.SetTag(tag);
      (PASN_Integer &)param2->m_parameterValue = 490; // level 3.1
       h264.m_collapsing.Append(param2);
       
        H245_GenericParameter * param3 = new H245_GenericParameter;
    param3->m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
       (PASN_Integer &)param3->m_parameterIdentifier = 4; //CustomMaxFS frame size parameter
       tag = H245_ParameterValue::e_unsignedMin;
        param3->m_parameterValue.SetTag(tag);
      (PASN_Integer &)param3->m_parameterValue = 36; //
       h264.m_collapsing.Append(param3);
       
        H245_GenericParameter * param4 = new H245_GenericParameter;
    param4->m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
       (PASN_Integer &)param4->m_parameterIdentifier = 5; //H.264 CustomMaxDPB memory parameter
       tag = H245_ParameterValue::e_unsignedMin;
        param4->m_parameterValue.SetTag(tag);
      (PASN_Integer &)param4->m_parameterValue = 106; // 
       h264.m_collapsing.Append(param4);
       
          H245_GenericParameter * param5 = new H245_GenericParameter;
    param5->m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
       (PASN_Integer &)param5->m_parameterIdentifier = 6; //H.264 CustomMaxDPB memory parameter
       tag = H245_ParameterValue::e_unsignedMin;
        param5->m_parameterValue.SetTag(tag);
      (PASN_Integer &)param5->m_parameterValue = 200; // 
       h264.m_collapsing.Append(param5);
       
          H245_GenericParameter * param6 = new H245_GenericParameter;
    param6->m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
       (PASN_Integer &)param6->m_parameterIdentifier = 7; //H.264 CustomMaxDPB memory parameter
       tag = H245_ParameterValue::e_unsignedMin;
        param6->m_parameterValue.SetTag(tag);
      (PASN_Integer &)param6->m_parameterValue = 490; // 
       h264.m_collapsing.Append(param6);
      
 

  return TRUE;
}


PBoolean H323_H264Capability::OnSendingPDU(H245_VideoMode & pdu) const
{
 
  return TRUE;
}


PBoolean H323_H264Capability::OnReceivedPDU(const H245_VideoCapability & cap)
{
if (cap.GetTag() != H245_VideoCapability::e_genericVideoCapability)
    return FALSE;

  const H245_GenericCapability & h264 = cap;

  maxBitRate = h264.m_maxBitRate;
  
  return TRUE;
}


H323Codec * H323_H264Capability::CreateCodec(H323Codec::Direction direction) const
{     
return new H323_H264Codec(direction, sqcifMPI, qcifMPI, cifMPI, cif4MPI, cif16MPI, maxBitRate, videoFrameRate);
}

//////////////////////////////////////////////////////////////////////////////


H323_H264Codec::H323_H264Codec(Direction dir,
			       unsigned _sqcifMPI,
			       unsigned _qcifMPI,
			       unsigned _cifMPI,
			       unsigned _cif4MPI,
			       unsigned _cif16MPI,
			       unsigned _maxBitRate,
			       unsigned _videoFrameRate)
  : H323VideoCodec("H.264", dir)
{
PTRACE(3, "H263\t" << (dir == Encoder ? "En" : "De") 
	 << "coder created. Data rate=" << _maxBitRate
	 << " Frame rate=" << _videoFrameRate);

  bitRateHighLimit = _maxBitRate;
  if (bitRateHighLimit == 0) {
    PTRACE(3, "H263\tData Rate is set to 1000 Kb/sec, as supplied value (0) is invalid");
    bitRateHighLimit = 1000 * 1024;
  }
  
  framesPerSec = _videoFrameRate;
  if (framesPerSec == 0) {
    framesPerSec = 25;
  }

  int shifts = -1;
  if (_sqcifMPI) { 
    shifts = 0;     PTRACE(3, "H263\t" << (dir == Encoder ? "En" : "De") << "coder for _sqcifMPI ");
  }
  if (_qcifMPI) {
    shifts = 1;  PTRACE(3, "H263\t" << (dir == Encoder ? "En" : "De") << "coder for _qcifMPI"); 
  }
  if (_cifMPI) { 
    shifts = 2; PTRACE(3, "H263\t" << (dir == Encoder ? "En" : "De") << "coder for _cifMPI");
  }
  if (_cif4MPI) {
    shifts = 3; PTRACE(3, "H263\t" << (dir == Encoder ? "En" : "De") << "coder for _cif4MPI");
  }
  if (_cif16MPI) {
    shifts = 4; PTRACE(3, "H263\t" << (dir == Encoder ? "En" : "De") << "coder for _cif16MPI");
  }

  if (shifts < 0) {
    PTRACE(1, "H263\tERROR in definition of h263 size");
    return;
  }
PTRACE(3, "H263\t" << (dir == Encoder ? "En" : "De") << "coder created." 
	 << "for a size of " << (88 << shifts) << "x" << (72 << shifts));
  
  Resize(88 << shifts , 72 << shifts); //Fill picture structure, open codec.
  frameNum = 0;

  InitialiseCodec();
}


H323_H264Codec::~H323_H264Codec()
{
  
}

void H323_H264Codec::InitialiseCodec()
{ 
   // no video decoder until we receive a packet
   videoDecoder = NULL;

   // no video encoder until we receive a packet
   videoEncoder = NULL;
}

void H323_H264Codec::CloseCodec()
{
  PTRACE(6, "H263\tClose h263 video " <<(direction == Encoder ? "En" : "De") << "coder");
}




/* Notes:
Quality was primarily for NetMeeting?
fillLevel simply set the greyscale of the background
*/


//This function grabs, displays, and compresses a video frame into
//into H263 packets.
//Get another frame if all packets of previous frame have been sent.
//Get next packet on list and send that one.
//Render the current frame if all of its packets have been sent.
PBoolean H323_H264Codec::Read(BYTE * buffer,
                          unsigned & length,
                          RTP_DataFrame & frame)
{
  return TRUE;
}




PBoolean H323_H264Codec::Write(const BYTE * buffer,
                           unsigned length,
                           const RTP_DataFrame & frame,
                           unsigned & written)
{


  return TRUE;
}




PBoolean H323_H264Codec::Resize(int _width, int _height)
{
  if ((frameWidth == _width) && (frameHeight == _height)) 
    return TRUE;

  PTRACE(6, "H263\t" << (direction == Encoder ? "En" : "De") << "coder resizing to "
	 << _width << "x" << _height << ".");


  frameWidth = _width;
  frameHeight = _height;

  return TRUE;
}
 

/* RenderFrame does two things:
   a) Set size of the display frame. This call happens with every frame.
	 A very small overhead.
   b) Display a frame.
*/

PBoolean H323_H264Codec::RenderFrame(const void * buffer)
{
  if (rawDataChannel == NULL)
    return TRUE;

  //Now display local image.
//  ((PVideoChannel *)rawDataChannel)->SetRenderFrameSize(frameWidth, frameHeight);

  if (buffer == NULL)
    return TRUE;

  return rawDataChannel->Write(buffer, 0 /*unused parameter*/);
}

/* AWM: Look-alike to H.261 implementation */
PBoolean H323_H264Codec::RenderFrame()
{
  void *srcData;
  

  return  TRUE ;//RenderFrame(srcData);
}




void H323_H264Codec::SetTxQualityLevel(int qLevel)
{
 /* int qualityLevel = PMIN(14, PMAX(qLevel,3));

  int lowLimit = PMIN(10, qualityLevel - 2);
  int highLimit = qualityLevel + 12;

  videoQuality = qLevel;
  videoQMin = lowLimit;       
  videoQMax = highLimit;*/
}


void H323_H264Codec::SetBackgroundFill(int idle)
{
 //   fillLevel = PMIN(99, PMAX(idle,1));
}


void H323_H264Codec::OnLostPartialPicture()
{
  PTRACE(3, "H263\tLost partial picture message ignored, not implemented");
}


void H323_H264Codec::OnLostPicture()
{
  PTRACE(3, "H263\tLost picture message ignored, not implemented");
}

 
//////////////////////////////////////////////////////////////////////

//#endif // H323_VICH263

